"""add_creator_in_vfolders

Revision ID: 4b7b650bc30e
Revises: 202b6dcbc159
Create Date: 2019-10-30 00:23:57.085692

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4b7b650bc30e"
down_revision = "202b6dcbc159"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("vfolders", sa.Column("creator", sa.String(length=128), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("vfolders", "creator")
    # ### end Alembic commands ###
