"""add is_local to images

Revision ID: 10c58e701d87
Revises: 36b69ddee76e
Create Date: 2023-02-24 06:28:41.583932

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "10c58e701d87"
down_revision = "36b69ddee76e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "images",
        sa.Column("is_local", sa.Boolean(), server_default=sa.text("false"), nullable=False),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("images", "is_local")
    # ### end Alembic commands ###
