"""add_resource_opts_in_kernels

Revision ID: 5b45f28d2cac
Revises: 9c89b9011872
Create Date: 2019-09-08 10:07:20.971662

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "5b45f28d2cac"
down_revision = "9c89b9011872"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "kernels",
        sa.Column("resource_opts", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "resource_opts")
    # ### end Alembic commands ###
